// WGPL - Wave++ General Purpose Library
// Copyright (C) 1995-2000 Wave++ (Yuri D'Elia)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// See the AUTHORS file for more info about the authors.  See the
// ChangeLog files for a list of changes.

// WGPL: definizione classi per i socket client/server

#include <wgpl_sockets.hpp>

// definizione classe TClientSocket
hostent TClientSocket::Tsock_gethostbyname(const char *host)
{
  struct hostent *he = gethostbyname(host);
  if (he == NULL) throw TClientSocket_gethost;
  return *he;
}

int TClientSocket::Tsock_newfd()
{
  int nfd = socket(AF_INET, SOCK_STREAM, 0);

  if (nfd == -1) throw TClientSocket_newfd;
  return nfd; 
}

struct sockaddr_in TClientSocket::Tsock_filladdr(const struct hostent &hst, const int &port)
{
  struct sockaddr_in hst_addr;

  hst_addr.sin_family = AF_INET;
  hst_addr.sin_port = htons(port);
  hst_addr.sin_addr = *((struct in_addr *)hst.h_addr);
  memset(&(hst_addr.sin_zero),0, 8);

  return hst_addr;
}

int TClientSocket::Tsock_connect(const int &fd, const struct sockaddr_in &addr)
{
  int rtval = ::connect(fd, (struct sockaddr *)&addr, sizeof(struct sockaddr));

  if (rtval == -1) throw TClientSocket_connect;
  return rtval;
}

int TClientSocket::Tsock_disconnect(const int &fd)
{
  int rtval = shutdown(fd,2);

  if (rtval == -1) throw TClientSocket_disconnect;
  return rtval;
}

double TClientSocket::Tsock_bps()
{
  struct timeval ctime = rsettimer();
  double transfbytes = sentbytes+recvbytes;
  double elapsedtime = ctime.tv_sec-timer.tv_sec;

  return (transfbytes/elapsedtime);
}

struct timeval TClientSocket::rsettimer()
{
  struct timeval ctime;
  gettimeofday(&ctime,NULL);
  return ctime;
}

TClientSocket::TClientSocket()
  : act(false) {}

TClientSocket::TClientSocket(const int &fd)
  : act(false)
{
  connect(fd);
}

TClientSocket::TClientSocket(const TClientSocket &source)
{
  act = source.act;
  
  if (act)
    {
      if (source.frmfd) throw TClientSocket_duplicatefromfd;

      sockfd = Tsock_newfd();
      hst_addr = source.hst_addr;
      Tsock_connect(sockfd,hst_addr);

      sentbytes = source.sentbytes;
      recvbytes = source.recvbytes;
      timer = source.timer;

      frmfd = false;
    }
}

TClientSocket::TClientSocket(const char *host, const int &port)
  : act(false)
{
  connect(host,port);
}

TClientSocket::~TClientSocket()
{
  if (!act || frmfd) return;

  Tsock_disconnect(sockfd);
}

int TClientSocket::getfd()
{
  frmfd = true;
  return sockfd;
}

void TClientSocket::connect(const char *host, const int &port)
{
  if (act && !frmfd) Tsock_disconnect(sockfd);

  hst_hent = Tsock_gethostbyname(host);
  sockfd = Tsock_newfd();
  hst_addr = Tsock_filladdr(hst_hent,port);
  Tsock_connect(sockfd,hst_addr);
  
  sentbytes = 0;
  recvbytes = 0;
  timer = rsettimer();

  frmfd = false;
  act = true;
}

void TClientSocket::connect(const int &fd)
{
  if (act && !frmfd) Tsock_disconnect(sockfd);

  sockfd = fd;

  sentbytes = 0;
  recvbytes = 0;
  timer = rsettimer();

  frmfd = true;
  act = true;
}

void TClientSocket::disconnect()
{
  if (!act || frmfd) return;

  act = false;
  Tsock_disconnect(sockfd);
}

int TClientSocket::send(const char *buf, const size_t &lenght)
{
  if (!act) return -1;
  
  int sbytes = ::send(sockfd,(void *)buf,lenght,0);
  if (sbytes == -1) throw TClientSocket_send;

  sentbytes+=lenght;
  return sbytes;
}

int TClientSocket::fsend(const char *buf, const size_t &sndlenght, const int &timeout = 0)
{
  size_t bts = sndlenght;
  size_t rval;
  char *tmpbuf = new char[sndlenght];

  while (bts > 0)
    {
      xstrncpy(tmpbuf,buf,0,sndlenght-bts,sndlenght-1);

      if (wsync(timeout)) break;
      rval = send(tmpbuf, bts);

      if (!rval) break;

      bts-=rval;
    }

  delete []tmpbuf;
  
  return sndlenght-bts;
}

int TClientSocket::urgsend(const char *buf, const size_t &lenght)
{
  if (!act) return -1;

  int sbytes = ::send(sockfd,(void *)buf,lenght,MSG_OOB);
  if (sbytes == -1) throw TClientSocket_urgsend;
  
  sentbytes+=lenght;
  return sbytes;
}

int TClientSocket::furgsend(const char *buf, const size_t &sndlenght, const int &timeout = 0)
{
  size_t bts = sndlenght;
  size_t rval;
  char *tmpbuf = new char[sndlenght];

  while (bts > 0)
    {
      xstrncpy(tmpbuf,buf,0,sndlenght-bts,sndlenght-1);

      if (wsync(timeout)) break;
      rval = urgsend(tmpbuf, bts);

      if (!rval) break;

      bts-=rval;
    }

  delete []tmpbuf;
  
  return sndlenght-bts;
}

int TClientSocket::read(char *buf, const size_t &maxlenght)
{
  if (!act) return -1;

  int rbytes = ::recv(sockfd,(void *)buf, maxlenght,0);
  if (rbytes == -1) throw TClientSocket_recv;

  recvbytes+=rbytes;
  return rbytes;
}

int TClientSocket::fread(char *buf, const size_t &recvlenght, const int &timeout = 0)
{
  size_t btr = recvlenght;
  size_t rval;
  char *tmpbuf = new char[recvlenght];

  while (btr > 0)
    {
      if (rsync(timeout)) break;
      rval = read(tmpbuf, btr);

      if (!rval) break;

      xstrncpy(buf,tmpbuf,recvlenght-btr,0,rval-1);

      btr-=rval;
    }

  delete []tmpbuf;
  
  return recvlenght-btr;
}

int TClientSocket::urgread(char *buf, const size_t &maxlenght)
{
  if (!act) return -1;

  int rbytes = ::recv(sockfd,(void *)buf, maxlenght,MSG_OOB);
  if (rbytes == -1) throw TClientSocket_urgrecv;

  recvbytes+=rbytes;
  return rbytes;
}

int TClientSocket::furgread(char *buf, const size_t &recvlenght, const int &timeout = 0)
{
  size_t btr = recvlenght;
  size_t rval;
  char *tmpbuf = new char[recvlenght];

  while (btr > 0)
    {
      if (xsync(timeout)) break;
      rval = urgread(tmpbuf, btr);

      if (!rval) break;

      xstrncpy(buf,tmpbuf,recvlenght-btr,0,rval-1);

      btr-=rval;
    }

  delete []tmpbuf;
  
  return recvlenght-btr;
}

char TClientSocket::sync(const int &timeout = 0)
{
  if (!act) return -1;

  fd_set readfds;
  fd_set writefds;
  fd_set exceptfds;
  int fds;
  char buf;
  char retval = 0x0;
  
  if (timeout)
    {
      hst_timeout.tv_sec = timeout;
      hst_timeout.tv_usec = 0;
    }
  
  FD_ZERO(&readfds);
  FD_ZERO(&writefds);
  FD_ZERO(&exceptfds);
  FD_SET(sockfd, &readfds);
  FD_SET(sockfd, &writefds);
  FD_SET(sockfd, &exceptfds);
  
  if (timeout)
    fds = select(sockfd+1, &readfds, &writefds, &exceptfds, &hst_timeout);
  else
    fds = select(sockfd+1, &readfds, &writefds, &exceptfds, NULL);

  if (fds == -1) throw TClientSocket_sync;

  if (timeout) { if (fds == 0) return TClientSocketSync::timeout; }

  if ((fds == 2) && (FD_ISSET(sockfd,&readfds)) && (FD_ISSET(sockfd,&writefds)))
    {
      if (recv(sockfd,(void *)&buf, 1,MSG_PEEK) == 0) retval |= TClientSocketSync::close;
    }

  if (FD_ISSET(sockfd,&readfds)) retval |= TClientSocketSync::read;
  if (FD_ISSET(sockfd,&writefds)) retval |= TClientSocketSync::write;
  if (FD_ISSET(sockfd,&exceptfds)) retval |= TClientSocketSync::exception;

  return retval;
}

bool TClientSocket::ateot()
{
  char buf;
  return (!recv(sockfd,(void *)&buf, 1,MSG_PEEK))?true:false;
}

bool TClientSocket::rsync(const int &timeout = 0)
{
  if (!act) return true;

  fd_set readfds;
  int fds;
  char buf;
  
  if (timeout)
    {
      hst_timeout.tv_sec = timeout;
      hst_timeout.tv_usec = 0;
    }
  
  FD_ZERO(&readfds);
  FD_SET(sockfd, &readfds);
  
  if (timeout)
    fds = select(sockfd+1, &readfds, NULL, NULL, &hst_timeout);
  else
    fds = select(sockfd+1, &readfds, NULL, NULL, NULL);

  if (fds == -1) throw TClientSocket_sync;

  if (timeout) { if (fds == 0) return true; }

  return false;
}

bool TClientSocket::wsync(const int &timeout = 0)
{
  if (!act) return true;

  fd_set writefds;
  int fds;
  char buf;
  
  if (timeout)
    {
      hst_timeout.tv_sec = timeout;
      hst_timeout.tv_usec = 0;
    }
  
  FD_ZERO(&writefds);
  FD_SET(sockfd, &writefds);
  
  if (timeout)
    fds = select(sockfd+1, NULL, &writefds, NULL, &hst_timeout);
  else
    fds = select(sockfd+1, NULL, &writefds, NULL, NULL);

  if (fds == -1) throw TClientSocket_sync;

  if (timeout) { if (fds == 0) return true; }

  return false;
}

bool TClientSocket::esync(const int &timeout = 0)
{
  if (!act) return true;

  fd_set exceptfds;
  int fds;
  char buf;
  
  if (timeout)
    {
      hst_timeout.tv_sec = timeout;
      hst_timeout.tv_usec = 0;
    }
  
  FD_ZERO(&exceptfds);
  FD_SET(sockfd, &exceptfds);
  
  if (timeout)
    fds = select(sockfd+1, NULL, NULL, &exceptfds, &hst_timeout);
  else
    fds = select(sockfd+1, NULL, NULL, &exceptfds, NULL);

  if (fds == -1) throw TClientSocket_sync;

  if (timeout) { if (fds == 0) return true; }

  return false;
}

char TClientSocket::sync(const char &event, const int &timeout = 0)
{
  if (!act) return -1;

  char retval;

  do retval = sync(timeout); while ((retval & event) == 0);

  return retval;
}

char TClientSocket::xsync(const char &event, const int &timeout = 0)
{
  if (!act) return -1;

  fd_set readfds;
  fd_set writefds;
  fd_set exceptfds;
  int fds;
  char buf;
  char retval = 0x0;
  
  if (timeout)
    {
      hst_timeout.tv_sec = timeout;
      hst_timeout.tv_usec = 0;
    }
  
  FD_ZERO(&readfds);
  FD_ZERO(&writefds);
  FD_ZERO(&exceptfds);
  if (((event & TClientSocketSync::read) > 0) ||
      ((event & TClientSocketSync::close) > 0)) FD_SET(sockfd, &readfds);
  if (((event & TClientSocketSync::write) > 0) ||
      ((event & TClientSocketSync::close) > 0)) FD_SET(sockfd, &writefds);
  if ((event & TClientSocketSync::exception) > 0) FD_SET(sockfd, &exceptfds);
  
  if (timeout)
    fds = select(sockfd+1, &readfds, &writefds, &exceptfds, &hst_timeout);
  else
    fds = select(sockfd+1, &readfds, &writefds, &exceptfds, NULL);

  if (fds == -1) throw TClientSocket_sync;

  if (timeout) { if (fds == 0) return TClientSocketSync::timeout; }

  if ((fds == 2) && (FD_ISSET(sockfd,&readfds)) &&
      (FD_ISSET(sockfd,&writefds)) && ((event & TClientSocketSync::close) > 0))
    {
      if (recv(sockfd,(void *)&buf, 1,MSG_PEEK) == 0) retval |= TClientSocketSync::close;
    }

  if (FD_ISSET(sockfd,&readfds)) retval |= TClientSocketSync::read;
  if (FD_ISSET(sockfd,&writefds)) retval |= TClientSocketSync::write;
  if (FD_ISSET(sockfd,&exceptfds)) retval |= TClientSocketSync::exception;

  return retval;
}

bool TClientSocket::connected()
{
  return act;
}

size_t TClientSocket::bytes_sent()
{
  return sentbytes;
}

size_t TClientSocket::bytes_recv()
{
  return recvbytes;
}

size_t TClientSocket::bytes_transfered()
{
  return sentbytes+recvbytes;
}

double TClientSocket::bps()
{
  return Tsock_bps();
}

double TClientSocket::kbps()
{
  return Tsock_bps()/1000;
}

TClientSocket &TClientSocket::operator=(const TClientSocket &source)
{
  if (this == &source) return *this;

  if (act) Tsock_disconnect(sockfd);

  act = source.act;
  
  if (act)
    {
      if (source.frmfd) throw TClientSocket_duplicatefromfd;

      sockfd = Tsock_newfd();
      hst_addr = source.hst_addr;
      Tsock_connect(sockfd,hst_addr);

      sentbytes = source.sentbytes;
      recvbytes = source.recvbytes;
      timer = source.timer;

      frmfd = false;
    }

  return *this;
}

bool TClientSocket::operator==(const TClientSocket &source)
{
  if ((!act) || (!source.act)) return false;

  if (memcmp((void *)&hst_addr,(void *)&source.hst_addr,sizeof(struct sockaddr_in))) return false;

  return true;
}

bool TClientSocket::operator==(const bool &source)
{
  return (act == source)?true:false;
}

bool TClientSocket::operator!()
{
  return !act;
}

bool TClientSocket::operator!=(const TClientSocket &source)
{
  return !operator==(source);
}

TClientSocket &TClientSocket::operator<<(const char *buf)
{
  if (!act) return *this;

  size_t lenght = strlen(buf);

  if (::send(sockfd,(void *)buf,lenght,0) == -1) throw TClientSocket_send;

  sentbytes+=lenght;
}

// definizione classe TClientSocket
int TServerSocket::Tsock_newfd()
{
  int nfd = socket(AF_INET, SOCK_STREAM, 0);

  if (nfd == -1) throw TServerSocket_newfd;
  return nfd; 
}

struct sockaddr_in TServerSocket::Tsock_filladdr(const int &port)
{
  struct sockaddr_in my_addr;

  my_addr.sin_family = AF_INET;
  my_addr.sin_port = htons(port);
  my_addr.sin_addr.s_addr = INADDR_ANY;
  bzero(&my_addr.sin_zero, 8);

  return my_addr;
}

int TServerSocket::Tsock_getport(const int &fd, const struct sockaddr_in &addr)
{
  struct sockaddr_in tmpaddr;
  size_t addrsize = sizeof(struct sockaddr);
  getsockname( fd, (struct sockaddr *)&tmpaddr, &addrsize);
  return ntohs(tmpaddr.sin_port);
}

int TServerSocket::Tsock_bind(const int &fd, const struct sockaddr_in &myaddr)
{
  int retval = bind(fd, (struct sockaddr *)&myaddr, sizeof(struct sockaddr));
  if (retval == -1) throw TServerSocket_bind;

  return retval;
}

int TServerSocket::Tsock_listen(const int &fd, const int &maxconn)
{
  int retval = listen(fd, maxconn);
  if (retval == -1) throw TServerSocket_listen;

  return retval;
}

TServerSocket::TServerSocket()
  : act(false) {}

TServerSocket::TServerSocket(const int &port, const int &maxconn)
  : act(false)
{
  start(port,maxconn);
}

int TServerSocket::Tsock_disconnect(const int &fd)
{  
  int rtval = close(fd);

  if (rtval == -1) throw TServerSocket_closefd;
  return rtval;
}

TServerSocket::~TServerSocket()
{
  if (!act) return;

  Tsock_disconnect(sockfd);
}

void TServerSocket::stop()
{
  if (!act) return;

  Tsock_disconnect(sockfd);
  act = false;
}

bool TServerSocket::active()
{
  return act;
}

int TServerSocket::getport()
{
  if (!act) return -1;

  return Tsock_getport(sockfd,hst_addr);
}

void TServerSocket::start(const int &port, const int &maxconn)
{
  if (act) Tsock_disconnect(sockfd);

  sockfd = Tsock_newfd();
  hst_addr = Tsock_filladdr(port);
  Tsock_bind(sockfd,hst_addr);
  Tsock_listen(sockfd,maxconn);
  act = true;
}

NewClientData TServerSocket::accept()
{
  NewClientData newfd;

  if (!act) { newfd.sockfd = -1; return newfd; };

  size_t sin_size = sizeof(struct sockaddr);
  newfd.sockfd = ::accept(sockfd, (struct sockaddr *)&rmt_addr, &sin_size);
  if (newfd.sockfd == -1) throw TServerSocket_accept;

  newfd.data = rmt_addr;
  return newfd;
}

bool TServerSocket::operator==(const TServerSocket &source)
{
  if ((!act) || (!source.act)) return false;
  
  if (memcmp((void *)&hst_addr,(void *)&source.hst_addr,sizeof(struct sockaddr_in))) return false;

  return true;
}

bool TServerSocket::operator==(const bool &source)
{
  return (act == source)?true:false;
}

bool TServerSocket::operator!=(const TServerSocket &source)
{
  return !operator==(source);
}

bool TServerSocket::operator!()
{
  return !act;
}
